# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import contextlib
import mimetypes
import pathlib
from typing import Annotated

import fastapi.responses
import rattler.exceptions
import rattler.platform

from mahoraga import _core

from . import _utils

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/{channel}/{platform}/{name}", dependencies=_core.immutable)
async def get_conda_package(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    name: Annotated[str, fastapi.Path(pattern=r"\.(?:conda|tar\.bz2)$")],
) -> fastapi.Response:
    return await _proxy_cache(channel, platform, name)


@router.get(
    "/{channel}/label/{label}/{platform}/{name}",
    dependencies=_core.immutable,
)
async def get_conda_package_with_label(
    channel: str,
    label: str,
    platform: rattler.platform.PlatformLiteral,
    name: Annotated[str, fastapi.Path(pattern=r"\.(?:conda|tar\.bz2)$")],
) -> fastapi.Response:
    return await _proxy_cache(channel, platform, name, label)


async def _proxy_cache(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    name: str,
    label: str | None = None,
) -> fastapi.Response:
    if name.endswith(".conda"):
        media_type = "application/zip"
        package_format_selection = rattler.PackageFormatSelection.ONLY_CONDA
        suffix = ".conda"
    else:
        media_type, _ = mimetypes.guess_type(name)
        package_format_selection = rattler.PackageFormatSelection.ONLY_TAR_BZ2
        suffix = ".tar.bz2"
    if label:
        cache_location = pathlib.Path(
            "channels", channel, "label", label, platform, name)
    else:
        cache_location = pathlib.Path("channels", channel, platform, name)
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(
                cache_location,
                media_type=media_type,
            )
        pkg_name, version, build = name.removesuffix(suffix).rsplit("-", 2)
        spec = f"{pkg_name} =={version}[{build=}]"
        try:
            records = await _utils.fetch_repo_data_and_load_matching_records(
                channel,
                platform,
                spec,
                package_format_selection,
                label=label,
            )
        except rattler.exceptions.FetchRepoDataError:
            return fastapi.Response(status_code=404)
        for record in records:
            if record.file_name == name:
                break
        else:
            return fastapi.Response(status_code=404)
        urls = _utils.urls(channel, platform, name, label)
        if record.sha256:
            return await _core.stream(
                urls,
                media_type=media_type,
                stack=stack,
                cache_location=cache_location,
                sha256=record.sha256,
                size=record.size,
            )
        return await _core.stream(
            urls,
            media_type=media_type,
            stack=stack,
        )
    return _core.unreachable()



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["make_app"]

import importlib.metadata
from typing import cast

import fastapi.openapi.docs
import fastapi.responses
import fastapi.templating
import jinja2
import starlette.middleware.cors
import starlette.staticfiles

from mahoraga import _conda, _core, _jsdelivr, _pypi, _python

URL_FOR = "{{ url_for('get_npm_file', package='swagger-ui-dist@5', path=%r) }}"


def make_app() -> fastapi.FastAPI:
    ctx = _core.context.get()
    cfg = ctx["config"]
    meta = importlib.metadata.metadata("mahoraga")
    contact = None
    if urls := meta.get_all("Project-URL"):
        for value in cast("list[str]", urls):
            if value.startswith("Issue Tracker, "):
                name, url = value.split(", ")
                contact = {"name": name, "url": url}
    app = fastapi.FastAPI(
        debug=cfg.log.level == "debug",
        title="Mahoraga",
        summary=meta["Summary"],
        version=meta["Version"],
        default_response_class=_JSONResponse,
        docs_url=None,
        redoc_url=None,
        contact=contact,
        license_info={
            "name": "License",
            "identifier": meta["License-Expression"],
        },
    )
    app.add_middleware(
        starlette.middleware.cors.CORSMiddleware,
        allow_origins=cfg.cors.allow_origins,
        allow_methods=cfg.cors.allow_methods,
        allow_headers=cfg.cors.allow_headers,
        allow_credentials=cfg.cors.allow_credentials,
        allow_origin_regex=cfg.cors.allow_origin_regex,
        expose_headers=cfg.cors.expose_headers,
        max_age=cfg.cors.max_age,
    )
    app.include_router(_conda.router, prefix="/conda", tags=["conda"])
    app.include_router(
        _conda.parselmouth,
        prefix="/parselmouth",
        tags=["conda"],
    )
    app.include_router(_jsdelivr.npm, prefix="/npm", tags=["pyodide"])
    app.include_router(_jsdelivr.pyodide, prefix="/pyodide", tags=["pyodide"])
    app.include_router(_pypi.router, prefix="/pypi", tags=["pypi"])
    app.include_router(_python.router, tags=["python"])
    app.mount(
        "/static",
        starlette.staticfiles.StaticFiles(packages=[("mahoraga", "_static")]),
        name="static",
    )

    # Private, only for building docs
    app.add_api_route(
        "/favicon.ico",
        _favicon,
        include_in_schema=False,
        response_class=fastapi.responses.RedirectResponse,
    )
    app.include_router(_jsdelivr.gh, prefix="/gh", include_in_schema=False)

    res = fastapi.openapi.docs.get_swagger_ui_html(
        openapi_url="{{ url_for('openapi') }}",
        title=app.title,
        swagger_js_url=URL_FOR % "swagger-ui-bundle.js",
        swagger_css_url=URL_FOR % "swagger-ui.css",
        swagger_favicon_url="{{ url_for('_favicon') }}",
        oauth2_redirect_url=URL_FOR % "oauth2-redirect.html",
        init_oauth=app.swagger_ui_init_oauth,
        swagger_ui_parameters=app.swagger_ui_parameters,
    )
    env = jinja2.Environment(autoescape=True)
    template = env.from_string(str(res.body, res.charset))
    name = cast("str", template)
    templates = fastapi.templating.Jinja2Templates(env=env)

    @app.get("/docs", include_in_schema=False)
    async def swagger_ui_html(
        request: fastapi.Request,
    ) -> fastapi.responses.HTMLResponse:
        return templates.TemplateResponse(request, name)

    del swagger_ui_html
    return app


class _JSONResponse(fastapi.responses.JSONResponse):
    media_type = None


async def _favicon(request: fastapi.Request) -> str:  # noqa: RUF029
    url = request.url_for(
        "get_scoped_npm_file",
        package="svg@0",
        path="filled/temple_buddhist.svg",
    )
    return str(url)



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "fetch_repo_data",
    "fetch_repo_data_and_load_matching_records",
    "prefix",
    "urls",
]

import itertools
import posixpath
from typing import TYPE_CHECKING

import rattler.exceptions
import rattler.networking
import rattler.platform
import rattler.rattler

from mahoraga import _core

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence

    from _typeshed import StrPath


async def fetch_repo_data(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    *,
    client: rattler.Client | None = None,
    label: str | None = None,
) -> rattler.SparseRepoData:
    if not client:
        ctx = _core.context.get()
        client = ctx["config"].server.rattler_client()
    if label:
        channel = f"{channel}/label/{label}"
    channels = [rattler.Channel(channel)]
    platforms = [rattler.Platform(platform)]
    try:
        [repodata] = await _fetch_repo_data(
            channels=channels,
            platforms=platforms,
            cache_path="repodata-cache",
            callback=None,
            client=client,
        )
    except rattler.exceptions.FetchRepoDataError:
        [repodata] = await rattler.fetch_repo_data(
            channels=channels,
            platforms=platforms,
            cache_path="repodata-cache",
            callback=None,
            fetch_options=_fetch_options,
        )
    return repodata


async def fetch_repo_data_and_load_matching_records(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    spec: str,
    package_format_selection: rattler.PackageFormatSelection,
    *,
    label: str | None = None,
) -> list[rattler.RepoDataRecord]:
    if label:
        channel = f"{channel}/label/{label}"
    channels = [rattler.Channel(channel)]
    platforms = [rattler.Platform(platform)]
    specs = [rattler.MatchSpec(spec, strict=True)]
    try:
        [repodata] = await rattler.fetch_repo_data(
            channels=channels,
            platforms=platforms,
            cache_path="repodata-cache",
            callback=None,
            fetch_options=_fetch_options,
        )
    except rattler.exceptions.FetchRepoDataError:
        pass
    else:
        with repodata:
            if records := repodata.load_matching_records(
                specs,
                package_format_selection,
            ):
                return records
    ctx = _core.context.get()
    [repodata] = await _fetch_repo_data(
        channels=channels,
        platforms=platforms,
        cache_path="repodata-cache",
        callback=None,
        client=ctx["config"].server.rattler_client(),
    )
    with repodata:
        return repodata.load_matching_records(specs, package_format_selection)


def prefix(channel: str) -> str:
    if channel == "emscripten-forge-dev":
        return "prefix.dev/emscripten-forge-dev"
    return f"conda.anaconda.org/{channel}"


def urls(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    name: str,
    label: str | None = None,
) -> list[str]:
    ctx = _core.context.get()
    cfg = ctx["config"].upstream.conda
    if label:
        return [
            posixpath.join(str(url), channel, "label", label, platform, name)
            for url in itertools.chain(
                cfg.default,
                _getitem(cfg.with_label, channel),
            )
        ]
    return [
        posixpath.join(str(url), channel, platform, name)
        for url in itertools.chain(
            cfg.default,
            _getitem(cfg.with_label, channel),
            _getitem(cfg.without_label, channel),
        )
    ]


async def _fetch_repo_data(  # noqa: PLR0913
    *,
    channels: list[rattler.Channel],
    platforms: list[rattler.Platform],
    cache_path: StrPath,
    callback: Callable[[int, int], None] | None,
    client: rattler.Client | None = None,
    fetch_options: rattler.networking.FetchRepoDataOptions | None = None,
) -> list[rattler.SparseRepoData]:
    fetch_options = fetch_options or rattler.networking.FetchRepoDataOptions()
    repo_data_list = await rattler.rattler.py_fetch_repo_data(  # pyright: ignore[]
        [channel._channel for channel in channels],  # noqa: SLF001  # pyright: ignore[]
        [platform._inner for platform in platforms],  # noqa: SLF001  # pyright: ignore[]
        cache_path,
        callback,
        client._client if client else None,  # noqa: SLF001  # pyright: ignore[]
        fetch_options._into_py(),  # noqa: SLF001  # pyright: ignore[]
    )
    return [
        rattler.SparseRepoData._from_py_sparse_repo_data(repo_data)  # noqa: SLF001  # pyright: ignore[]
        for repo_data in repo_data_list  # pyright: ignore[reportUnknownVariableType]
    ]


def _getitem[T](mapping: Mapping[str, str | list[T]], key: str) -> Sequence[T]:
    seen = {key}
    value = mapping.get(key, ())
    while isinstance(value, str):
        if value in seen:
            raise RecursionError
        seen.add(value)
        value = mapping.get(value, ())
    return value


_fetch_options = rattler.networking.FetchRepoDataOptions(
    cache_action="force-cache-only",
)



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["GitHubRelease", "NPMBase"]

import asyncio
import os
import pathlib
from typing import TYPE_CHECKING, Any, Literal, Self

import pooch  # pyright: ignore[reportMissingTypeStubs]
import pydantic

from mahoraga import _core

if TYPE_CHECKING:
    from _typeshed import StrPath


class _ReleaseAsset(pydantic.BaseModel, extra="ignore"):
    url: str
    name: str
    size: int
    digest: str | None


class GitHubRelease(pydantic.BaseModel, extra="ignore"):
    assets: list[_ReleaseAsset]

    @classmethod
    async def fetch(
        cls,
        *args: StrPath,
        owner: str,
        repo: str,
        tag_name: str,
    ) -> Self:
        if not args:
            _core.unreachable()
        cache_location = pathlib.Path(*args)
        ctx = _core.context.get()
        async with ctx["locks"][str(cache_location)]:
            return await asyncio.to_thread(
                _fetch,
                cls,
                f"https://api.github.com/repos/{owner}/{repo}/releases/tags/{tag_name}",
                cache_location,
                headers=_headers,
            )


class NPMBase(pydantic.BaseModel):
    type: Literal["npm"]
    name: str
    version: str

    @classmethod
    async def fetch(cls, *args: StrPath, url: str, **kwargs: object) -> Self:
        if not args:
            _core.unreachable()
        cache_location = pathlib.Path(*args)
        ctx = _core.context.get()
        async with ctx["locks"][str(cache_location)]:
            return await asyncio.to_thread(
                _fetch, cls, url, cache_location, **kwargs)


def _fetch[T: pydantic.BaseModel](
    klass: type[T],
    url: str,
    cache_location: pathlib.Path,
    **kwargs: Any,  # noqa: ANN401
) -> T:
    path, fname = os.path.split(cache_location)
    pooch.retrieve(  # pyright: ignore[reportUnknownMemberType]
        url,
        known_hash=None,
        fname=fname,
        path=path,
        downloader=pooch.HTTPDownloader(**kwargs),
    )
    json_data = cache_location.read_text(encoding="utf-8")
    try:
        return klass.model_validate_json(json_data)
    except pydantic.ValidationError:
        cache_location.unlink()
        raise


_headers = {
    "Accept": "application/vnd.github+json",
    "X-GitHub-Api-Version": "2022-11-28",
}



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import asyncio
import contextlib
import contextvars
from typing import Annotated

import fastapi

from mahoraga import _core

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/compressed_mapping.json")
async def get_compressed_mapping() -> fastapi.Response:
    ctx = contextvars.copy_context()
    lock = ctx[_core.context]["locks"]["compressed_mapping.json"]
    ctx.run(_core.cache_action.set, "cache-or-fetch")
    async with contextlib.AsyncExitStack() as stack:
        await stack.enter_async_context(lock)
        return await asyncio.create_task(
            _core.stream(
                "https://api.github.com/repos/prefix-dev/parselmouth/contents/files/compressed_mapping.json",
                headers={
                    "Accept": "application/vnd.github.raw+json",
                    "X-GitHub-Api-Version": "2022-11-28",
                },
                stack=stack,
            ),
            context=ctx,
        )


@router.get("/hash-v0/{sha256}")
async def get_hash_mapping(
    sha256: Annotated[str, fastapi.Path(pattern=r"^[0-9a-f]{64}$")],
) -> fastapi.Response:
    return fastapi.Response(status_code=404)



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["NormalizedName", "Simple"]

import functools
from typing import Annotated

import annotated_types as at
import packaging.utils
import pydantic

NormalizedName = Annotated[
    str,
    pydantic.AfterValidator(
        functools.partial(packaging.utils.canonicalize_name, validate=True),
    ),
]


class _Hashes(pydantic.BaseModel, extra="ignore"):
    sha256: str = ""


class _Entry(pydantic.BaseModel, extra="ignore"):
    filename: str
    hashes: _Hashes

    # Not required, see https://peps.python.org/pep-0691/#project-detail
    size: int | None = None


class _Meta(pydantic.BaseModel, extra="ignore"):
    api_version: Annotated[
        str,
        at.Ge("1.0"),
        at.Lt("1.:"),
        pydantic.Field(alias="api-version"),
    ]


class Simple(pydantic.BaseModel, extra="ignore"):
    name: NormalizedName
    files: list[_Entry]
    meta: _Meta



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import asyncio
import contextlib
import contextvars
import http
import mimetypes
import pathlib
import posixpath
from typing import TYPE_CHECKING, Annotated

import fastapi.responses
import httpx
import packaging.utils

from mahoraga import _core

from . import _models

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.head("/{tag}/{prefix}/{project}/{filename}")
async def check_pypi_package_availability(
    tag: str,
    prefix: Annotated[str, fastapi.Path(min_length=1, max_length=2)],
    project: str,
    filename: str,
) -> fastapi.Response:
    ctx = _core.context.get()
    client = ctx["httpx_client"]
    try:
        response = await client.head(
            f"https://files.pythonhosted.org/packages/{tag}/{prefix}/{project}/{filename}",
        )
    except httpx.HTTPError:
        return fastapi.Response()
    if response.has_redirect_location:
        return fastapi.Response()
    return _core.Response(
        response.content,
        response.status_code,
        response.headers,
    )


@router.get(
    "/{tag}/{prefix}/{project}/{filename}",
    dependencies=_core.immutable,
)
async def get_pypi_package(
    tag: str,
    prefix: Annotated[str, fastapi.Path(min_length=1, max_length=2)],
    project: str,
    filename: str,
) -> fastapi.Response:
    if filename.endswith(".metadata"):
        return await _core.stream(
            f"https://files.pythonhosted.org/packages/{tag}/{prefix}/{project}/{filename}",
        )
    if filename.endswith(".whl"):
        normalized_name = packaging.utils.parse_wheel_filename(filename)[0]
        media_type = "application/zip"
    else:
        normalized_name, _ = packaging.utils.parse_sdist_filename(filename)
        media_type, _ = mimetypes.guess_type(filename)
    cache_location = pathlib.Path("packages", tag, prefix, project, filename)
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(
                cache_location,
                media_type=media_type,
            )
        ctx = _core.context.get()
        match len(tag), len(prefix), len(project):
            case (2, 2, 60):
                urls = [
                    posixpath.join(str(url), "packages", tag, prefix,
                                   project, filename)
                    for url in ctx["config"].upstream.pypi.all()
                ]
            case (_, 1, _) if project.startswith(prefix):
                client = ctx["httpx_client"]
                try:
                    response = await stack.enter_async_context(
                        client.stream(
                            "GET",
                            f"https://files.pythonhosted.org/packages/{tag}/{prefix}/{project}/{filename}",
                        ),
                    )
                except httpx.HTTPError as e:
                    raise fastapi.HTTPException(
                        http.HTTPStatus.GATEWAY_TIMEOUT,
                    ) from e
                if not response.has_redirect_location:
                    new_stack = stack.pop_all()
                    content = _stream(response, new_stack)
                    try:
                        await anext(content)
                    except:
                        stack.push_async_exit(new_stack)
                        raise
                    return _core.StreamingResponse(
                        content,
                        response.status_code,
                        response.headers,
                    )
                _core.schedule_exit(stack)
                p = httpx.URL(response.headers["Location"]).path.lstrip("/")
                urls = [
                    posixpath.join(str(url), p)
                    for url in ctx["config"].upstream.pypi.all()
                ]
            case _:
                raise fastapi.HTTPException(404)
        sha256, size = await _sha256(filename, normalized_name)
        if sha256:
            return await _core.stream(
                urls,
                media_type=media_type,
                stack=stack,
                cache_location=cache_location,
                sha256=sha256,
                size=size,
            )
        return await _core.stream(
            urls,
            media_type=media_type,
            stack=stack,
        )
    return _core.unreachable()


async def _sha256(filename: str, project: str) -> tuple[bytes, int | None]:
    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    match ctx[_core.context]:
        case {"config": config, "locks": locks}:
            pass
        case _:
            _core.unreachable()
    urls = [
        posixpath.join(str(url), "simple", project) + "/"
        for url in config.upstream.pypi.json_
    ]
    ctx.run(_core.cache_action.set, "force-cache-only")
    lock = locks[f"{project}|application/vnd.pypi.simple.v1+html"]
    try:
        async with lock:
            raw = await loop.create_task(
                _core.get(
                    urls,
                    headers={"Accept": "application/vnd.pypi.simple.v1+html"},
                ),
                context=ctx,
            )
    except (NotImplementedError, fastapi.HTTPException):
        pass
    else:
        return (
            await loop.run_in_executor(None, _sha256_from_html, raw, filename),
            None,
        )
    ctx.run(_core.cache_action.set, "cache-or-fetch")
    try:
        async with locks[f"{project}|application/vnd.pypi.simple.v1+json"]:
            raw = await loop.create_task(
                _core.get(
                    urls,
                    headers={"Accept": "application/vnd.pypi.simple.v1+json"},
                ),
                context=ctx,
            )
    except fastapi.HTTPException:
        pass
    else:
        return await loop.run_in_executor(
            None,
            _sha256_and_size_from_json,
            raw,
            project,
            filename,
        )
    urls += [
        posixpath.join(str(url), "simple", project) + "/"
        for url in config.upstream.pypi.html
    ]
    try:
        async with lock:
            raw = await loop.create_task(
                _core.get(
                    urls,
                    headers={"Accept": "application/vnd.pypi.simple.v1+html"},
                ),
                context=ctx,
            )
    except fastapi.HTTPException:
        return b"", None
    return (
        await loop.run_in_executor(None, _sha256_from_html, raw, filename),
        None,
    )


def _sha256_from_html(raw: bytes, filename: str) -> bytes:
    pattern = b"/%b#sha256=" % filename.encode("ascii")
    try:
        i = raw.index(pattern)
    except ValueError:
        return b""
    i += len(pattern)
    return bytes.fromhex(raw[i : i+64])  # noqa: E226  # pyright: ignore[reportArgumentType]


def _sha256_and_size_from_json(
    raw: bytes,
    project: str,
    filename: str,
) -> tuple[bytes, int | None]:
    simple = _models.Simple.model_validate_json(raw)
    if simple.name != project:
        _core.unreachable()
    for entry in simple.files:
        if entry.filename == filename:
            return bytes.fromhex(entry.hashes.sha256), entry.size
    return b"", None


async def _stream(
    response: httpx.Response,
    stack: contextlib.AsyncExitStack,
) -> AsyncGenerator[bytes]:
    async with stack:
        yield b""
        async for chunk in response.aiter_bytes():
            yield chunk



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Client-side patch for `panel` to work along with `mahoraga`."""

import dataclasses
import importlib.abc
import importlib.util
import sys
import types  # noqa: TC003
from importlib.machinery import ModuleSpec  # noqa: TC003
from typing import override


def monkey_patch(cdn: str = "http://127.0.0.1:3450/") -> None:
    """Patch `panel` to use a custom CDN.

    Raises:
        RuntimeError: `panel` has been imported before patching.

    """
    for name in "panel", "panel.io":
        if name in sys.modules:
            message = "%r has been imported"
            raise RuntimeError(message % name)
        sys.modules[name] = importlib.util.module_from_spec(_find_spec(name))
    name = "panel.io.resources"
    try:
        spec = _find_spec(name)
    finally:
        del sys.modules["panel.io"], sys.modules["panel"]
    spec.loader = loader = _PatchLoader(spec.loader, cdn)
    sys.modules[name] = module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)


@importlib.util.LazyLoader.factory  # pyright: ignore[reportArgumentType]
@dataclasses.dataclass
class _PatchLoader(importlib.abc.Loader):
    loader: importlib.abc.InspectLoader
    cdn: str

    @override
    def create_module(self, spec: ModuleSpec) -> types.ModuleType | None:
        return self.loader.create_module(spec)

    @override
    def exec_module(self, module: types.ModuleType) -> None:
        code = self.loader.get_code(module.__name__)
        if not code:
            message = "cannot load module %r when get_code() returns None"
            raise ImportError(message % module.__name__)
        consts = tuple(
            f"{self.cdn.rstrip('/')}/npm/@holoviz/panel@"
            if x == "https://cdn.holoviz.org/panel/" else x
            for x in code.co_consts
        )
        exec(code.replace(co_consts=consts), module.__dict__)  # noqa: S102


def _find_spec(name: str) -> ModuleSpec:
    if spec := importlib.util.find_spec(name):
        return spec
    message = "No module named %r"
    raise ModuleNotFoundError(message % name, name=name)



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import asyncio
import base64
import contextlib
import http
import logging
import mimetypes
import os
import pathlib
import posixpath
from typing import Annotated, Literal

import fastapi.responses
import packaging.version
import pooch  # pyright: ignore[reportMissingTypeStubs]
import pyodide_lock

from mahoraga import _core, _jsdelivr

from . import _utils

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/{name}", dependencies=_core.immutable)
async def get_pyodide_package(
    name: Annotated[
        str,
        fastapi.Path(
            pattern=r"^(pyodide|pyodide-core|static-libraries|xbuildenv)-.+\.tar\.bz2$",
        ),
    ],
) -> fastapi.Response:
    version = name[:-8].rsplit("-", 1)[1]
    try:
        packaging.version.parse(version)
    except packaging.version.InvalidVersion:
        return fastapi.Response(status_code=404)
    cache_location = pathlib.Path("pyodide", name)
    media_type, _ = mimetypes.guess_type(name)
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(
                cache_location,
                media_type=media_type,
            )
        release = await _core.GitHubRelease.fetch(
            f"pyodide/{version}.json",
            owner="pyodide",
            repo="pyodide",
            tag_name=version,
        )
        for asset in release.assets:
            if asset.name == name:
                break
        else:
            return fastapi.Response(status_code=404)
        headers = {
            "Accept": "application/octet-stream",
            "X-GitHub-Api-Version": "2022-11-28",
        }
        if digest := asset.digest:
            if digest.startswith("sha256:"):
                return await _core.stream(
                    asset.url,
                    headers=headers,
                    media_type=media_type,
                    stack=stack,
                    cache_location=cache_location,
                    sha256=bytes.fromhex(digest[7:]),
                    size=asset.size,
                )
            _logger.warning("GitHub returning non-SHA256 digest: %r", digest)
        return await _core.stream(asset.url, headers=headers, stack=stack)
    return _core.unreachable()


@router.get("/dev/{build}/{path:path}")
async def get_pyodide_dev_file(
    build: Literal["full", "debug"],
    path: Annotated[str, fastapi.Path(pattern=r"^[^/]")],
) -> fastapi.Response:
    urls = _utils.urls("pyodide", "dev", build, path)
    media_type, _ = mimetypes.guess_type(path)
    # Never cached
    return await _core.stream(urls, media_type=media_type)


@router.get("/{version}/full/package.json", dependencies=_core.immutable)
@router.get("/{version}/full/pyodide.asm.js", dependencies=_core.immutable)
@router.get("/{version}/full/pyodide.asm.wasm", dependencies=_core.immutable)
@router.get("/{version}/full/pyodide.js", dependencies=_core.immutable)
@router.get("/{version}/full/pyodide.js.map", dependencies=_core.immutable)
@router.get("/{version}/full/pyodide.mjs", dependencies=_core.immutable)
@router.get("/{version}/full/pyodide.mjs.map", dependencies=_core.immutable)
@router.get("/{version}/full/pyodide-lock.json", dependencies=_core.immutable)
@router.get("/{version}/full/python_stdlib.zip", dependencies=_core.immutable)
@router.get("/{version}/debug/package.json", dependencies=_core.immutable)
@router.get("/{version}/debug/pyodide.js.map", dependencies=_core.immutable)
@router.get("/{version}/debug/pyodide.mjs", dependencies=_core.immutable)
@router.get("/{version}/debug/pyodide.mjs.map", dependencies=_core.immutable)
@router.get("/{version}/debug/pyodide-lock.json", dependencies=_core.immutable)
@router.get("/{version}/debug/python_stdlib.zip", dependencies=_core.immutable)
async def get_pyodide_core_file(
    version: Annotated[str, fastapi.Path(pattern=r"^v")],
    request: fastapi.Request,
) -> fastapi.Response:
    version = version.lstrip("v")
    package = f"pyodide@{version}"
    name = posixpath.basename(request.url.path)
    cache_location = pathlib.Path("npm", package, name)
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(cache_location)
        return await _utils.get_npm_file(
            f"https://data.jsdelivr.com/v1/packages/npm/{package}",
            package,
            name,
            cache_location,
            stack,
        )
    return _core.unreachable()


@router.get(
    "/{version}/full/python_cli_entry.mjs",
    dependencies=_core.immutable,
)
@router.get(
    "/{version}/debug/python_cli_entry.mjs",
    dependencies=_core.immutable,
)
async def get_python_cli_entry(
    version: Annotated[str, fastapi.Path(pattern=r"^v")],
) -> fastapi.Response:
    member = "python_cli_entry.mjs"
    cache_location = pathlib.Path("pyodide", version, "full", member)
    async with _core.cached_or_locked(cache_location) as cached:
        if cached:
            return fastapi.responses.FileResponse(cache_location)
        for name in "pyodide-core", "xbuildenv", "pyodide":
            tarball = pathlib.Path("pyodide", f"{name}-{version[1:]}.tar.bz2")
            if response := await _utils.extract_from_tarball(
                tarball,
                member,
                cache_location,
            ):
                return response
    urls = _utils.urls("pyodide", version, "full", member)
    return await _core.stream(urls, media_type="text/javascript")


@router.get("/{version}/{build}/{path:path}", dependencies=_core.immutable)
async def get_pyodide_file(
    version: Annotated[str, fastapi.Path(pattern=r"^v")],
    build: Literal["full", "debug"],
    path: Annotated[str, fastapi.Path(pattern=r"^[^/]")],
) -> fastapi.Response:
    urls = _utils.urls("pyodide", version, build, path)
    media_type, _ = mimetypes.guess_type(path)
    if "/" in path or not path.endswith((".whl", ".zip", "-tests.tar")):
        # No way to find the checksums
        return await _core.stream(urls, media_type=media_type)
    cache_location = pathlib.Path("pyodide", version, "full", path)
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(cache_location)
        version = version.lstrip("v")
        tarball = pathlib.Path("pyodide", f"pyodide-{version}.tar.bz2")
        if response := await _utils.extract_from_tarball(
            tarball,
            path,
            cache_location,
        ):
            return response
        package = f"pyodide@{version}"
        json_file = pathlib.Path("npm", package, "pyodide-lock.json")
        await _get_pyodide_lock(json_file, package)
        for spec in (
            pyodide_lock.PyodideLockSpec.from_json(json_file).packages.values()
        ):
            if spec.file_name == path:
                break
        else:
            return fastapi.Response(status_code=404)
        if path.endswith(".whl"):
            media_type = "application/zip"
            urls.append(
                f"https://anaconda.org/pyodide/{spec.name}/{spec.version}/download/{path}",
            )
        return await _core.stream(
            urls,
            media_type=media_type,
            stack=stack,
            cache_location=cache_location,
            sha256=bytes.fromhex(spec.sha256),
        )
    return _core.unreachable()


async def _get_pyodide_lock(
    cache_location: pathlib.Path,
    package: str,
) -> None:
    async with _core.cached_or_locked(cache_location) as cached:
        if not cached:
            for name in "pyodide-core", "xbuildenv":
                tarball = pathlib.Path(
                    "pyodide",
                    f"{name}-{package[8:]}.tar.bz2",
                )
                if await _utils.extract_from_tarball(
                    tarball,
                    "pyodide-lock.json",
                    cache_location,
                ):
                    return
            metadata = await _jsdelivr.Metadata.fetch(
                f"npm/{package}.json",
                url=f"https://data.jsdelivr.com/v1/packages/npm/{package}",
                params={"structure": "flat"},
            )
            for file in metadata.files:
                if file["name"] == "/pyodide-lock.json":
                    break
            else:
                raise fastapi.HTTPException(404)
            known_hash = base64.b64decode(file["hash"]).hex()
            dir_, fname = os.path.split(cache_location)
            loop = asyncio.get_running_loop()
            urls = _utils.urls("npm", package, "pyodide-lock.json")
            async with contextlib.aclosing(_core.load_balance(urls)) as it:
                async for url in it:
                    with contextlib.suppress(Exception):
                        await loop.run_in_executor(
                            None,
                            pooch.retrieve,  # pyright: ignore[reportUnknownArgumentType, reportUnknownMemberType]
                            url,
                            known_hash,
                            fname,
                            dir_,
                        )
                        break
                else:
                    status_code = http.HTTPStatus.GATEWAY_TIMEOUT
                    raise fastapi.HTTPException(status_code)


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router", "split_repo"]

import asyncio
import compression.zstd
import hashlib
import json
import logging
import pathlib
import shutil
from typing import Annotated, Any

import fastapi.responses
import msgpack
import pooch.utils  # pyright: ignore[reportMissingTypeStubs]
import rattler.platform

from mahoraga import _core

from . import _models, _utils

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get(
    "/{channel}/{platform}/repodata_shards.msgpack.zst",
    dependencies=_core.hourly,
)
async def get_sharded_repodata_index(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
) -> fastapi.Response:
    return fastapi.responses.FileResponse(
        f"channels/{channel}/{platform}/repodata_shards.msgpack.zst",
    )


@router.get(
    "/{channel}/label/{label}/{platform}/repodata_shards.msgpack.zst",
    dependencies=_core.hourly,
)
async def get_sharded_repodata_index_with_label(
    channel: str,
    label: str,
    platform: rattler.platform.PlatformLiteral,
) -> fastapi.Response:
    return fastapi.responses.FileResponse(
        f"channels/{channel}/label/{label}/{platform}/repodata_shards.msgpack.zst",
    )


@router.get(
    "/{channel}/{platform}/shards/{name}",
    dependencies=_core.immutable,
)
async def get_sharded_repodata(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    name: Annotated[str, fastapi.Path(pattern=r"\.msgpack\.zst$")],
) -> fastapi.Response:
    return fastapi.responses.FileResponse(
        f"channels/{channel}/{platform}/shards/{name}",
    )


@router.get(
    "/{channel}/label/{label}/{platform}/shards/{name}",
    dependencies=_core.immutable,
)
async def get_sharded_repodata_with_label(
    channel: str,
    label: str,
    platform: rattler.platform.PlatformLiteral,
    name: Annotated[str, fastapi.Path(pattern=r"\.msgpack\.zst$")],
) -> fastapi.Response:
    return fastapi.responses.FileResponse(
        f"channels/{channel}/label/{label}/{platform}/shards/{name}",
    )


async def split_repo() -> None:  # noqa: RUF029
    ctx = _core.context.get()
    cfg = ctx["config"]
    if any(cfg.shard.values()):
        executor = ctx["process_pool"]
        for channel, platforms in cfg.shard.items():
            for platform in platforms:
                executor.submit(_worker, cfg, channel, platform)


def _packages(
    package_name: rattler.PackageName,
    package_format_selection: rattler.PackageFormatSelection,
    repodata: rattler.SparseRepoData,
    run_exports: dict[str, dict[str, Any]],
) -> dict[str, dict[str, Any]]:
    shards: dict[str, dict[str, Any]] = {}
    for record in repodata.load_records(
        package_name,
        package_format_selection,
    ):
        old = _models.PackageRecord.model_validate_json(record.to_json())
        if new := old.__pydantic_extra__:
            if old.md5:
                new["md5"] = bytes.fromhex(old.md5)
            if old.sha256:
                new["sha256"] = bytes.fromhex(old.sha256)
            filename = record.file_name
            if entry := run_exports.get(filename):
                new["run_exports"] = entry["run_exports"]
            shards[filename] = new
    return shards


def _sha256(
    name: str,
    repodata: rattler.SparseRepoData,
    root: pathlib.Path,
    run_exports: _models.RunExports,
) -> bytes:
    package_name = rattler.PackageName(name)
    shard: _models.Shard = {
        "packages": _packages(
            package_name,
            rattler.PackageFormatSelection.ONLY_TAR_BZ2,
            repodata,
            run_exports["packages"],
        ),
        "packages.conda": _packages(
            package_name,
            rattler.PackageFormatSelection.ONLY_CONDA,
            repodata,
            run_exports["packages.conda"],
        ),
        "removed": [],
    }
    with pooch.utils.temporary_file(root) as tmp:  # pyright: ignore[reportUnknownMemberType]
        with pathlib.Path(tmp).open("w+b") as f:
            with compression.zstd.ZstdFile(f, "w", level=19) as g:
                msgpack.dump(shard, g)
            f.seek(0)
            h = hashlib.file_digest(f, "sha256")
        dst = root / f"{h.hexdigest()}.msgpack.zst"
        dst.unlink(missing_ok=True)
        shutil.move(tmp, dst)
    return h.digest()


def _split_repo(
    cfg: _core.Config,
    channel: str,
    platform: rattler.platform.PlatformLiteral,
) -> None:
    root = pathlib.Path("channels", channel, platform, "shards")
    root.mkdir(parents=True, exist_ok=True)
    json_file = root.with_name("run_exports.json.zst")
    try:
        new = pooch.retrieve(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
            f"https://{_utils.prefix(channel)}/{platform}/run_exports.json.zst",
            known_hash=None,
            path=root.parent,
        )
    except OSError:
        pass
    else:
        json_file.unlink(missing_ok=True)
        shutil.move(new, json_file)
    try:
        f = compression.zstd.ZstdFile(json_file)
    except OSError:
        run_exports: _models.RunExports = {
            "packages": {},
            "packages.conda": {},
        }
    else:
        with f:
            run_exports = json.load(f)
    with asyncio.run(
        _utils.fetch_repo_data(
            channel,
            platform,
            client=cfg.server.rattler_client(),
        ),
        debug=cfg.log.level == "debug",
        loop_factory=cfg.loop_factory,
    ) as repodata:
        sharded_repodata: _models.ShardedRepodata = {
            "info": {
                "base_url": ".",
                "shards_base_url": "./shards/",
                "subdir": platform,
            },
            "shards": {
                name: _sha256(name, repodata, root, run_exports)
                for name in repodata.package_names()
            },
        }
    dst = root.with_name("repodata_shards.msgpack.zst")
    with pooch.utils.temporary_file(root.parent) as tmp:  # pyright: ignore[reportUnknownMemberType]
        with compression.zstd.ZstdFile(tmp, "w", level=19) as f:
            msgpack.dump(sharded_repodata, f)
        dst.unlink(missing_ok=True)
        shutil.move(tmp, dst)


def _worker(
    cfg: _core.Config,
    channel: str,
    platform: rattler.platform.PlatformLiteral,
) -> None:
    try:
        _split_repo(cfg, channel, platform)
    except Exception:
        _logger.exception(
            "Failed to update %s/%s/repodata_shards.msgpack.zst",
            channel,
            platform,
        )
    else:
        _logger.info(
            "Successfully updated %s/%s/repodata_shards.msgpack.zst",
            channel,
            platform,
        )


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import asyncio
import contextlib
import http
import logging
import mimetypes
import os
import pathlib
import posixpath
import re
import urllib.parse
from typing import Annotated

import fastapi.responses
import packaging.version
import pooch  # pyright: ignore[reportMissingTypeStubs]
import pydantic
import pydantic_extra_types.semantic_version  # noqa: TC002
import requests

from . import _core

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/python/pymanager/{name}")
async def get_python_install_manager_for_windows(
    name: Annotated[
        str,
        fastapi.Path(pattern=r"^python-manager-.+\.msix?$"),
    ],
) -> fastapi.Response:
    if not re.fullmatch(
        packaging.version.VERSION_PATTERN,
        name[15 : name.endswith(".msi") - 5],
        re.VERBOSE | re.IGNORECASE,
    ):
        return fastapi.Response(status_code=404)
    ctx = _core.context.get()
    urls = [
        urllib.parse.unquote(str(url)).format(version="pymanager", name=name)
        for url in ctx["config"].upstream.python
    ]
    return await _core.stream(urls)


@router.get("/python/{version}/{name}")
async def get_embedded_python(
    version: pydantic_extra_types.semantic_version.SemanticVersion,
    name: str,
) -> fastapi.Response:
    if (
        version < "3.5.0"
        or version.prerelease is not None
        or version.build is not None
        or name not in {
            f"python-{version}-embed-amd64.zip",
            f"python-{version}-embed-arm64.zip",
            f"python-{version}-embed-win32.zip",
        }
    ):
        return fastapi.Response(status_code=404)
    ctx = _core.context.get()
    urls = [
        urllib.parse.unquote(str(url)).format(version=version, name=name)
        for url in ctx["config"].upstream.python
    ]
    media_type, _ = mimetypes.guess_type(name)
    return await _core.stream(urls, media_type=media_type)


@router.get(
    "/python-build-standalone/{tag}/{name}",
    dependencies=_core.immutable,
)
async def get_standalone_python(
    tag: Annotated[
        str,
        fastapi.Path(min_length=8, max_length=8, pattern=r"^\d+$"),
    ],
    name: str,
) -> fastapi.Response:
    media_type, _ = mimetypes.guess_type(name)
    cache_location = pathlib.Path("python-build-standalone", tag, name)
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(
                cache_location,
                media_type=media_type,
            )
        ctx = _core.context.get()
        escaped = name.replace("+", "%2B")  # Required by aliyun mirrors
        urls = [
            posixpath.join(str(url), tag, escaped)
            for url in ctx["config"].upstream.python_build_standalone
        ]
        sha256, size = await _get_standalone_python_sha256_and_size(tag, name)
        return await _core.stream(
            urls,
            media_type=media_type,
            stack=stack,
            cache_location=cache_location,
            sha256=sha256,
            size=size,
        )
    return _core.unreachable()


async def _get_standalone_python_github_release(
    tag: str,
    name: str,
) -> tuple[bytes, int | None]:
    release = await _core.GitHubRelease.fetch(
        f"python-build-standalone/{tag}.json",
        owner="astral-sh",
        repo="python-build-standalone",
        tag_name=tag,
    )
    for asset in release.assets:
        if asset.name == name:
            break
    else:
        raise requests.RequestException
    if digest := asset.digest:
        if digest.startswith("sha256:"):
            return bytes.fromhex(digest[7:]), asset.size
        _logger.warning("GitHub returning non-SHA256 digest: %r", digest)
    raise requests.RequestException


async def _get_standalone_python_sha256_and_size(
    tag: str,
    name: str,
) -> tuple[bytes, int | None]:
    try:
        return await _get_standalone_python_github_release(tag, name)
    except requests.RequestException:
        pass
    except (OSError, pydantic.ValidationError):
        _logger.exception("Failed to get GitHub release metadata")
    cache_location = pathlib.Path("python-build-standalone", tag, "SHA256SUMS")
    loop = asyncio.get_running_loop()
    async with _core.cached_or_locked(cache_location) as cached:
        if not cached:
            ctx = _core.context.get()
            dir_, fname = os.path.split(cache_location)
            urls = [
                posixpath.join(str(url), tag, "SHA256SUMS")
                for url in ctx["config"].upstream.python_build_standalone
            ]
            async with contextlib.aclosing(_core.load_balance(urls)) as it:
                async for url in it:
                    with contextlib.suppress(Exception):
                        await loop.run_in_executor(
                            None,
                            pooch.retrieve,  # pyright: ignore[reportUnknownArgumentType, reportUnknownMemberType]
                            url,
                            None,
                            fname,
                            dir_,
                        )
                        break
                else:
                    status_code = http.HTTPStatus.GATEWAY_TIMEOUT
                    raise fastapi.HTTPException(status_code)
    return await loop.run_in_executor(
        None,
        _parse_standalone_python_sha256,
        cache_location,
        f"  {name}\n".encode("ascii"),
    )


def _parse_standalone_python_sha256(
    cache_location: pathlib.Path,
    name: bytes,
) -> tuple[bytes, None]:
    with cache_location.open("rb") as f:
        for line in f:
            if line.endswith(name):
                return bytes.fromhex(line[:64]), None  # pyright: ignore[reportArgumentType]
    raise fastapi.HTTPException(404)


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "APIRoute",
    "Response",
    "StreamingResponse",
    "get",
    "load_balance",
    "stream",
]

import asyncio
import contextlib
import hashlib
import http
import logging
import pathlib
import shutil
from collections.abc import (
    AsyncGenerator,
    Callable,
    Coroutine,
    Generator,
    Iterable,
    Mapping,
)
from typing import TYPE_CHECKING, Any, TypedDict, Unpack, overload, override

import anyio
import fastapi.responses
import fastapi.routing
import hishel.fastapi
import httpx
import pooch.utils  # pyright: ignore[reportMissingTypeStubs]

from mahoraga import _core

if TYPE_CHECKING:
    from _hashlib import HASH
    from types import TracebackType

    from _typeshed import ReadableBuffer, StrPath

hourly = [
    hishel.fastapi.cache(max_age=3600),
]
immutable = [
    hishel.fastapi.cache(max_age=31536000, public=True, immutable=True),
]


class APIRoute(fastapi.routing.APIRoute):
    @override
    def get_route_handler(self) -> Callable[
        [fastapi.Request],
        Coroutine[Any, Any, fastapi.Response],
    ]:
        async def wrapper(request: fastapi.Request) -> fastapi.Response:
            match response := await wrapped(request):
                case fastapi.responses.FileResponse():
                    _get_stack(request).push(_wrap_file_not_found_error)
                case fastapi.responses.StreamingResponse(
                    body_iterator=AsyncGenerator() as agen,
                ):
                    _get_stack(request).push_async_callback(agen.aclose)
                case _:
                    pass
            return response
        wrapped = super().get_route_handler()
        return wrapper


class Response(fastapi.Response):
    @override
    def init_headers(self, headers: Mapping[str, str] | None = None) -> None:
        headers = httpx.Headers(headers)
        for key in "Content-Encoding", "Date", "Server", "Transfer-Encoding":
            headers.pop(key, None)
        if self.media_type != type(self).media_type:
            headers.pop("Content-Type", None)
        super().init_headers(headers)


class StreamingResponse(fastapi.responses.StreamingResponse, Response):
    media_type = "application/octet-stream"


async def get(urls: Iterable[str], **kwargs: object) -> bytes:
    ctx = _core.context.get()
    client = ctx["httpx_client"]
    response = None
    async with (
        contextlib.AsyncExitStack() as stack,
        contextlib.aclosing(load_balance(urls)) as it,
    ):
        async for url in it:
            try:
                response = await stack.enter_async_context(
                    client.stream("GET", url, **kwargs),
                )
            except httpx.HTTPError:
                continue
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError:
                _core.schedule_exit(stack)
                continue
            try:
                return await response.aread()
            except httpx.StreamError:
                _core.schedule_exit(stack)
    if not response:
        raise fastapi.HTTPException(http.HTTPStatus.GATEWAY_TIMEOUT)
    headers = response.headers
    for key in "Date", "Server":
        headers.pop(key, None)
    raise fastapi.HTTPException(response.status_code, headers=dict(headers))


class _CacheOptions(TypedDict, total=False):
    cache_location: StrPath | None
    sha256: bytes | None
    size: int | None


@overload
async def stream(
    urls: Iterable[str],
    *,
    headers: Mapping[str, str] | None = ...,
    media_type: str | None = ...,
    stack: contextlib.AsyncExitStack | None = None,
    cache_location: None = ...,
    sha256: None = ...,
    size: int | None = ...,
) -> fastapi.Response: ...

@overload
async def stream(
    urls: Iterable[str],
    *,
    headers: Mapping[str, str] | None = ...,
    media_type: str | None = ...,
    stack: contextlib.AsyncExitStack | None = None,
    cache_location: StrPath,
    sha256: bytes,
    size: int | None = ...,
) -> fastapi.Response: ...


async def stream(
    urls: Iterable[str],
    *,
    headers: Mapping[str, str] | None = None,
    media_type: str | None = None,
    stack: contextlib.AsyncExitStack | None = None,
    **kwargs: Unpack[_CacheOptions],
) -> fastapi.Response:
    if stack:
        return await _entered(
            stack,
            urls,
            headers=headers,
            media_type=media_type,
            **kwargs,
        )
    async with contextlib.AsyncExitStack() as new_stack:
        return await _entered(
            new_stack,
            urls,
            headers=headers,
            media_type=media_type,
            **kwargs,
        )
    return _core.unreachable()


class _ContentLengthError(Exception):
    pass


async def _entered(
    stack: contextlib.AsyncExitStack,
    urls: Iterable[str],
    *,
    headers: Mapping[str, str] | None = None,
    media_type: str | None = None,
    **kwargs: Unpack[_CacheOptions],
) -> fastapi.Response:
    ctx = _core.context.get()
    client = ctx["httpx_client"]
    inner_stack = await stack.enter_async_context(contextlib.AsyncExitStack())
    response = None
    async with contextlib.aclosing(load_balance(urls)) as it:
        async for url in it:
            try:
                response = await inner_stack.enter_async_context(
                    client.stream("GET", url, headers=headers),
                )
            except httpx.HTTPError:
                continue
            if response.status_code == http.HTTPStatus.NOT_MODIFIED:
                break
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError:
                _core.schedule_exit(inner_stack)
                continue
            try:
                headers = _unify_content_length(response.headers, kwargs)
            except _ContentLengthError:
                _core.schedule_exit(inner_stack)
                response = None
                continue
            new_stack = contextlib.AsyncExitStack()
            content = _stream(response, new_stack, **kwargs)
            try:
                if await anext(content):
                    _core.unreachable()
            except httpx.TransportError:
                _core.schedule_exit(inner_stack)
                response = None
                continue
            await new_stack.enter_async_context(stack.pop_all())
            return StreamingResponse(
                content,
                response.status_code,
                headers,
                media_type,
            )
    if response:
        return Response(
            status_code=response.status_code,
            headers=response.headers,
        )
    return fastapi.Response(status_code=http.HTTPStatus.GATEWAY_TIMEOUT)


async def load_balance(urls: Iterable[str]) -> AsyncGenerator[str]:
    if isinstance(urls, str):
        urls = {urls}
    else:
        ctx = _core.context.get()
        lock = ctx["locks"]["statistics.json"]
        urls = set(urls)
        while len(urls) > 1:
            async with lock:
                url = min(urls, key=_core.Statistics().key)
            urls.remove(url)
            yield url
    for url in urls:
        yield url


def _get_stack(request: fastapi.Request) -> contextlib.AsyncExitStack:
    stack: contextlib.AsyncExitStack
    match request.scope:
        case {"fastapi_inner_astack": contextlib.AsyncExitStack() as stack}:  # pyright: ignore[reportUnknownVariableType]
            return stack
        case _:
            return _core.unreachable()


async def _stream(
    response: httpx.Response,
    wrapped: contextlib.AsyncExitStack,
    *,
    cache_location: StrPath | None = None,
    sha256: bytes | None = None,
    size: int | None = None,
) -> AsyncGenerator[bytes]:
    last = b""
    scope = anyio.CancelScope(shield=True)
    async with contextlib.AsyncExitStack() as stack:
        outer = await stack.enter_async_context(contextlib.AsyncExitStack())
        await stack.enter_async_context(wrapped)
        if cache_location and sha256:
            h = hashlib.sha256()
            inner = contextlib.ExitStack()
            loop = asyncio.get_running_loop()

            @stack.callback
            def _() -> None:
                try:
                    outer.enter_context(scope)
                finally:
                    fut = loop.run_in_executor(None, inner.close)
                    outer.push_async_callback(lambda: fut)
            fwrite = await loop.run_in_executor(
                None,
                inner.enter_context,
                _tempfile(response, cache_location, sha256, size, h),
            )
            async for current in response.aiter_bytes():
                fut = loop.run_in_executor(None, fwrite, current)
                yield last
                last = current
                await fut
                h.update(current)
        else:
            stack.callback(outer.enter_context, scope)
            if cache_location or sha256:
                _core.unreachable()
            async for current in response.aiter_bytes():
                yield last
                last = current
        yield last


@contextlib.contextmanager
def _tempfile(
    response: httpx.Response,
    cache_location: StrPath,
    sha256: bytes,
    size: int | None,
    hash_: HASH,
) -> Generator[Callable[[ReadableBuffer], int]]:
    dir_ = pathlib.Path(cache_location).parent
    dir_.mkdir(parents=True, exist_ok=True)
    with (
        pooch.utils.temporary_file(dir_) as tmp,  # pyright: ignore[reportUnknownMemberType]
        contextlib.ExitStack() as stack,
        pathlib.Path(tmp).open("wb") as f,
    ):
        f.truncate(size)
        try:
            yield f.write
        finally:
            if hash_.digest() == sha256 and (
                size is None
                or size == (
                    f.tell()
                    if "Content-Encoding" in response.headers
                    else response.num_bytes_downloaded
                )
            ):
                stack.callback(shutil.move, tmp, cache_location)


def _unify_content_length(
    headers: httpx.Headers,
    kwargs: _CacheOptions,
) -> httpx.Headers:
    if "Content-Encoding" in headers:
        # Content-Length refer to the encoded data, see
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Reference/Headers/Content-Encoding
        headers = headers.copy()
        if size := kwargs.get("size"):
            headers["Content-Length"] = str(size)
        else:
            headers.pop("Content-Length", None)
        # Content-Encoding will be removed in Response.init_headers
        return headers
    if content_length := headers.get("Content-Length"):
        actual = int(content_length)
        expect = kwargs.setdefault("size", actual)
        if expect is None:
            kwargs["size"] = actual
        elif expect != actual:
            _logger.warning(
                "Content-Length mismatch: expect %d, got %d",
                expect,
                actual,
            )
            raise _ContentLengthError
    return headers


def _wrap_file_not_found_error(
    _exc_type: type[BaseException] | None,
    exc_value: BaseException | None,
    _traceback: TracebackType | None,
) -> None:
    match exc_value:
        case RuntimeError(__context__=FileNotFoundError() as e):
            raise fastapi.HTTPException(404) from e
        case _:
            pass


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import mimetypes
import posixpath
from typing import TYPE_CHECKING, Annotated

import fastapi
import httpx
import rattler.platform  # noqa: TC002

from mahoraga import _core

from . import _models, _utils

if TYPE_CHECKING:
    from pydantic import BaseModel

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.head("/{channel}/{platform}/repodata.json.bz2")
@router.head("/{channel}/{platform}/repodata.json.zst")
@router.head("/{channel}/{platform}/repodata.jlap")
async def check_repodata_availability(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    request: fastapi.Request,
) -> fastapi.Response:
    return await _check_repodata_availability(channel, platform, request)


@router.head("/{channel}/label/{label}/{platform}/repodata.json.bz2")
@router.head("/{channel}/label/{label}/{platform}/repodata.json.zst")
@router.head("/{channel}/label/{label}/{platform}/repodata.jlap")
async def check_repodata_availability_with_label(
    channel: str,
    label: str,
    platform: rattler.platform.PlatformLiteral,
    request: fastapi.Request,
) -> fastapi.Response:
    channel = f"{channel}/label/{label}"
    return await _check_repodata_availability(channel, platform, request)


@router.get("/{channel}/{platform}/repodata.json")
@router.get("/{channel}/{platform}/repodata.json.bz2")
@router.get("/{channel}/{platform}/repodata.json.zst")
async def get_repodata(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    request: fastapi.Request,
    headers: Annotated[_models.RepodataHeaders, fastapi.Header()],
) -> fastapi.Response:
    return await _get_repodata(channel, platform, request, headers)


@router.get("/{channel}/label/{label}/{platform}/repodata.json")
@router.get("/{channel}/label/{label}/{platform}/repodata.json.bz2")
@router.get("/{channel}/label/{label}/{platform}/repodata.json.zst")
async def get_repodata_with_label(
    channel: str,
    label: str,
    platform: rattler.platform.PlatformLiteral,
    request: fastapi.Request,
    headers: Annotated[_models.RepodataHeaders, fastapi.Header()],
) -> fastapi.Response:
    return await _get_repodata(channel, platform, request, headers, label)


@router.get("/{channel}/{platform}/repodata.jlap")
async def get_differential_repodata(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    headers: Annotated[_models.JLAPHeaders, fastapi.Header()],
) -> fastapi.Response:
    return await _core.stream(
        f"https://{_utils.prefix(channel)}/{platform}/repodata.jlap",
        headers=_to_dict(headers),
    )


@router.get("/{channel}/label/{label}/{platform}/repodata.jlap")
async def get_differential_repodata_with_label(
    channel: str,
    label: str,
    platform: rattler.platform.PlatformLiteral,
    headers: Annotated[_models.JLAPHeaders, fastapi.Header()],
) -> fastapi.Response:
    return await _core.stream(
        f"https://conda.anaconda.org/{channel}/label/{label}/{platform}/repodata.jlap",
        headers=_to_dict(headers),
    )


async def _check_repodata_availability(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    request: fastapi.Request,
) -> fastapi.Response:
    ctx = _core.context.get()
    client = ctx["httpx_client"]
    name = posixpath.basename(request.url.path)
    try:
        response = await client.head(
            f"https://{_utils.prefix(channel)}/{platform}/{name}",
            follow_redirects=True,
        )
    except httpx.HTTPError:
        return fastapi.Response()
    return _core.Response(
        response.content,
        response.status_code,
        response.headers,
    )


async def _get_repodata(
    channel: str,
    platform: rattler.platform.PlatformLiteral,
    request: fastapi.Request,
    headers: _models.RepodataHeaders,
    label: str | None = None,
) -> fastapi.Response:
    name = posixpath.basename(request.url.path)
    urls = _utils.urls(channel, platform, name, label)
    media_type, _ = mimetypes.guess_type(name)
    return await _core.stream(
        urls,
        headers=_to_dict(headers),
        media_type=media_type,
    )


def _to_dict(obj: BaseModel) -> dict[str, str]:
    return {
        k.replace("_", "-"): v
        for k, v in obj.model_dump(exclude_none=True).items()
    }



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import mimetypes
from typing import Literal

import fastapi

from mahoraga import _core

from . import _utils

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/pyodide/{version}/docs/_static/img/pyodide-logo-readme.png")
async def pyodide_logo(version: str) -> fastapi.Response:
    if not version.startswith("pyodide@"):
        return fastapi.Response(status_code=404)
    urls = _utils.urls(
        "gh/pyodide",
        version,
        "docs/_static/img/pyodide-logo-readme.png",
    )
    return await _core.stream(urls, media_type="image/png")


@router.get("/jdecked/{version}/assets/{fmt}/{name}")
async def twemoji(
    version: str,
    fmt: Literal["svg", "72x72"],
    name: str,
) -> fastapi.Response:
    if not version.startswith("twemoji@"):
        return fastapi.Response(status_code=404)
    urls = _utils.urls("gh/jdecked", version, "assets", fmt, name)
    media_type, _ = mimetypes.guess_type(name)
    return await _core.stream(urls, media_type=media_type)



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["Address", "Config", "Predicate", "Server"]

import asyncio
import functools
import ipaddress
import itertools
import sys
from typing import TYPE_CHECKING, Annotated, Literal, no_type_check, override

import annotated_types as at
import pydantic
import pydantic_settings
import rattler.networking
import rattler.platform

if TYPE_CHECKING:
    from collections.abc import Iterator

if sys.platform == "win32":
    import winloop as uvloop  # pyright: ignore[reportMissingImports]
else:
    import uvloop  # pyright: ignore[reportMissingImports]

Predicate = functools.singledispatch(at.Predicate)


@Predicate.register
def _(func: str) -> at.Predicate:
    def wrapper(input_value: object, /) -> bool:
        return eval(func, {"input_value": input_value})  # noqa: S307
    wrapper.__qualname__ = func
    return at.Predicate(wrapper)


_model_config: pydantic.ConfigDict = {
    "alias_generator": lambda x: x.replace("_", "-"),
}


class Address(pydantic.BaseModel):
    host: Annotated[
        ipaddress.IPv4Address,
        pydantic.Field(description="The host to serve on"),
        Predicate("not (input_value.is_multicast or input_value.is_reserved)"),
    ] = ipaddress.IPv4Address("127.0.0.1")
    port: Annotated[
        pydantic.PositiveInt,
        pydantic.Field(description="The TCP port to serve on"),
        at.Le(65535),
    ] = 3450


class Server(Address, **_model_config):
    limit_concurrency: Annotated[
        int,
        pydantic.Field(description="Maximum number of simultaneous connections"
                                   " to allow"),
        at.Ge(2),
    ] = 512
    backlog: Annotated[
        pydantic.PositiveInt,
        pydantic.Field(description="Maximum number of pending connections to "
                                   "allow"),
    ] = 511
    keep_alive: Annotated[
        pydantic.PositiveInt,
        pydantic.Field(description="Time in seconds to wait before closing "
                                   "idle connections"),
    ] = 5
    timeout_graceful_shutdown: Annotated[
        pydantic.NonNegativeInt,
        pydantic.Field(description="Time in seconds to wait before graceful "
                                   "shutdown"),
    ] = 0

    def rattler_client(self) -> rattler.Client:
        return _rattler_client(self.host, self.port)


class _HttpUrl(pydantic.AnyUrl):
    _constraints = pydantic.UrlConstraints(
        max_length=1900,
        allowed_schemes=["http", "https"],
        host_required=True,
    )


class _Log(pydantic.BaseModel):
    level: Literal["debug", "info", "warning", "error", "critical"] = "info"
    access: bool = True


class _CORS(pydantic.BaseModel, **_model_config):
    allow_origins: list[str] = []
    allow_methods: list[str] = ["GET"]
    allow_headers: list[str] = []
    allow_credentials: bool = False
    allow_origin_regex: str | None = None
    expose_headers: list[str] = []
    max_age: pydantic.NonNegativeInt = 600


_adapter = pydantic.TypeAdapter(list[_HttpUrl])


class _Conda(pydantic.BaseModel, **_model_config):
    default: list[_HttpUrl] = _adapter.validate_python([
        "https://conda.anaconda.org/",
    ])
    with_label: dict[str, str | list[_HttpUrl]] = {
        "nvidia": _adapter.validate_python([
            "https://mirrors.sustech.edu.cn/anaconda-extra/cloud/",
        ]),
    }
    without_label: dict[str, str | list[_HttpUrl]] = {
        "auto": _adapter.validate_python([
            "https://mirror.nju.edu.cn/anaconda/cloud/",
            "https://mirrors.cqupt.edu.cn/anaconda/cloud/",
            "https://mirrors.hit.edu.cn/anaconda/cloud/",
            "https://mirrors.lzu.edu.cn/anaconda/cloud/",
            "https://mirrors.pku.edu.cn/anaconda/cloud/",
            "https://mirrors.sustech.edu.cn/anaconda/cloud/",
            "https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud/",
        ]),
        "biobakery": "auto",
        "bioconda": _adapter.validate_python([
            "https://mirror.nju.edu.cn/anaconda/cloud/",
            "https://mirrors.cqupt.edu.cn/anaconda/cloud/",
            "https://mirrors.hit.edu.cn/anaconda/cloud/",
            "https://mirrors.lzu.edu.cn/anaconda/cloud/",
            "https://mirrors.pku.edu.cn/anaconda/cloud/",
            "https://mirrors.sustech.edu.cn/anaconda/cloud/",
            "https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud/",
            "https://prefix.dev/",
        ]),
        "c4aarch64": "auto",
        "caffe2": "auto",
        "conda-forge": "bioconda",
        "deepmodeling": "auto",
        "dglteam": "auto",
        "emscripten-forge-dev": "nvidia",
        "fastai": "auto",
        "fermi": "auto",
        "idaholab": "auto",
        "matsci": "auto",
        "menpo": "auto",
        "MindSpore": "auto",
        "mordred-descriptor": "auto",
        "msys2": "auto",
        "numba": "auto",
        "nvidia": _adapter.validate_python([
            "https://prefix.dev/",
        ]),
        "ohmeta": "auto",
        "omnia": "auto",
        "Paddle": "auto",
        "peterjc123": "auto",
        "plotly": "auto",
        "psi4": "auto",
        "pytorch": "bioconda",
        "pytorch-lts": "auto",
        "pytorch-test": "auto",
        "pytorch3d": "auto",
        "pyviz": "auto",
        "qiime2": "auto",
        "rapidsai": "bioconda",
        "rdkit": "auto",
        "robostack": "nvidia",
        "robostack-humble": "nvidia",
        "robostack-jazzy": "nvidia",
        "robostack-noetic": "nvidia",
        "robostack-staging": "nvidia",
        "simpleitk": "auto",
        "speleo3": "auto",
        "stackless": "auto",
        "ursky": "auto",
    }


class _PyPI(pydantic.BaseModel):
    html: list[_HttpUrl] = _adapter.validate_python([
        "https://mirror.nju.edu.cn/pypi/web/",
        "https://mirrors.aliyun.com/pypi/web/",
        "https://mirrors.cloud.tencent.com/pypi/",
        "https://mirrors.huaweicloud.com/repository/pypi/",
        "https://mirrors.neusoft.edu.cn/pypi/web/",
        "https://mirrors.pku.edu.cn/pypi/web/",
        "https://mirrors.sustech.edu.cn/pypi/web/",
    ])
    json_: Annotated[list[_HttpUrl], pydantic.Field(alias="json")] = (
        _adapter.validate_python([
            "https://mirrors.bfsu.edu.cn/pypi/web/",
            "https://mirrors.tuna.tsinghua.edu.cn/pypi/web/",
            "https://pypi.org/",
        ])
    )

    def all(self) -> Iterator[_HttpUrl]:
        return itertools.chain(self.html, self.json_)


class _Upstream(pydantic.BaseModel, **_model_config):
    conda: _Conda = _Conda()
    pyodide: list[_HttpUrl] = _adapter.validate_python([
        "https://cdn.jsdelivr.net/",
        "https://fastly.jsdelivr.net/",
        "https://gcore.jsdelivr.net/",
        "https://originfastly.jsdelivr.net/",
        "https://quantil.jsdelivr.net/",
        "https://testingcf.jsdelivr.net/",
    ])
    pypi: _PyPI = _PyPI()
    python: list[_HttpUrl] = _adapter.validate_python([
        "https://cdn.npmmirror.com/binaries/python/{version}/{name}",
        "https://mirror.nju.edu.cn/python/{version}/{name}",
        "https://mirrors.aliyun.com/python-release/windows/{name}",
        "https://mirrors.bfsu.edu.cn/python/{version}/{name}",
        "https://mirrors.huaweicloud.com/python/{version}/{name}",
        "https://mirrors.tuna.tsinghua.edu.cn/python/{version}/{name}",
        "https://mirrors.ustc.edu.cn/python/{version}/{name}",
        "https://www.python.org/ftp/python/{version}/{name}",
    ])
    python_build_standalone: list[_HttpUrl] = _adapter.validate_python([
        "https://mirror.nju.edu.cn/github-release/astral-sh/python-build-standalone/",
        "https://mirrors.lzu.edu.cn/github-release/astral-sh/python-build-standalone/",
        "https://cdn.npmmirror.com/binaries/python-build-standalone/",
        "https://pycdn-2025-03-02.oss-cn-shanghai.aliyuncs.com/mirror/astral-sh/python-build-standalone/",
        "https://github.com/astral-sh/python-build-standalone/releases/download/",
    ])
    backup: set[str] = {
        "conda.anaconda.org",
        "github.com",
        "prefix.dev",
        "pycdn-2025-03-02.oss-cn-shanghai.aliyuncs.com",
        "pypi.org",
        "www.python.org",
    }


class Config(
    pydantic_settings.BaseSettings,
    toml_file="mahoraga.toml",
    **_model_config,
):
    server: Server = Server()
    log: _Log = _Log()
    shard: dict[str, set[rattler.platform.PlatformLiteral]] = {}
    cors: _CORS = _CORS()
    upstream: _Upstream = _Upstream()
    eager_task_execution: bool = False

    @no_type_check
    def loop_factory(self) -> asyncio.AbstractEventLoop:
        loop = uvloop.new_event_loop()
        if self.eager_task_execution:
            loop.set_task_factory(asyncio.eager_task_factory)
        return loop

    @override
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[pydantic_settings.BaseSettings],
        init_settings: pydantic_settings.PydanticBaseSettingsSource,
        env_settings: pydantic_settings.PydanticBaseSettingsSource,
        dotenv_settings: pydantic_settings.PydanticBaseSettingsSource,
        file_secret_settings: pydantic_settings.PydanticBaseSettingsSource,
    ) -> tuple[pydantic_settings.PydanticBaseSettingsSource, ...]:
        return (
            pydantic_settings.TomlConfigSettingsSource(settings_cls),
            init_settings,
        )


@functools.lru_cache(maxsize=1)
def _rattler_client(host: ipaddress.IPv4Address, port: int) -> rattler.Client:
    if host.is_unspecified:
        host = ipaddress.IPv4Address("127.0.0.1")
    return rattler.Client([
        rattler.networking.MirrorMiddleware({
            "https://conda.anaconda.org/": [
                str(_HttpUrl(f"http://{host}:{port}/conda/")),
            ],
        }),
    ])



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["run"]

import contextlib
from typing import TYPE_CHECKING

import filelock

from . import _server

if TYPE_CHECKING:
    from _typeshed import StrPath


def run(root: StrPath) -> None:
    with (
        contextlib.chdir(root),
        filelock.FileLock("mahoraga.lock", timeout=0, thread_local=False),
    ):
        _server.run()



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["Metadata", "router"]

import contextlib
import logging
import pathlib
from typing import Annotated, TypedDict

import fastapi.responses

from mahoraga import _core

from . import _utils

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/@bokeh/{package}/{path:path}", dependencies=_core.immutable)
@router.get("/@holoviz/{package}/{path:path}", dependencies=_core.immutable)
@router.get("/@pyscript/{package}/{path:path}", dependencies=_core.immutable)
@router.get("/@stlite/{package}/{path:path}", dependencies=_core.immutable)
# Must be the last
@router.get(
    "/@material-design-icons/{package}/{path:path}",
    dependencies=_core.immutable,
)
async def get_scoped_npm_file(
    package: Annotated[str, fastapi.Path(pattern=r"^[^@]+@[^@]+$")],
    path: Annotated[str, fastapi.Path(pattern=r"^[^/]")],
    request: fastapi.Request,
) -> fastapi.Response:
    return await get_npm_file(package, path, request)


@router.get("/{package}/{path:path}", dependencies=_core.immutable)
async def get_npm_file(
    package: Annotated[str, fastapi.Path(pattern=r"^[^@]+@[^@]+$")],
    path: Annotated[str, fastapi.Path(pattern=r"^[^/]")],
    request: fastapi.Request,
) -> fastapi.Response:
    cache_location = pathlib.Path(request.url.path.lstrip("/"))
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(cache_location)
        prefix = request.url.path.removeprefix("/npm")
        if prefix.startswith("/@"):
            scope, _ = prefix[2:].split("/", 1)
        elif package.startswith(("pyodide@", "swagger-ui-dist@")):
            scope = None
        else:
            return fastapi.Response(status_code=404)
        package, version = package.split("@")
        if scope:
            package = f"@{scope}/{package}"
        resolved = await _Resolved.fetch(
            "npm", package, f"{version}.json",
            url=f"https://data.jsdelivr.com/v1/packages/npm/{package}/resolved",
            params={"specifier": version},
        )
        package = f"{package}@{resolved.version}"
        if resolved.version != version:
            cache_location = pathlib.Path("npm", package, path)
            if await _core.cached_or_locked(cache_location, stack):
                return fastapi.responses.FileResponse(cache_location)
        return await _utils.get_npm_file(
            resolved.links["self"],
            package,
            path,
            cache_location,
            stack,
        )
    return _core.unreachable()


class _File(TypedDict):
    name: str
    hash: str
    size: int


class _MetadataLinks(TypedDict):
    entrypoints: str
    stats: str


class Metadata(_core.NPMBase):
    default: str | None
    files: list[_File]
    links: _MetadataLinks


class _ResolvedLinks(_MetadataLinks):
    self: str


class _Resolved(_core.NPMBase):
    links: _ResolvedLinks


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["main"]

import argparse
import contextlib
import pathlib
import sys
import urllib.parse
from typing import Annotated

import click
import jinja2
import pydantic
import pydantic_settings

from mahoraga import __version__, _asgi, _core


def main() -> None:
    """CLI entry."""
    cli_settings_source = (
        pydantic_settings.CliSettingsSource[argparse.ArgumentParser](_Main)
    )
    cli_settings_source.root_parser.suggest_on_error = True
    pydantic_settings.CliApp.run(
        _Main,
        cli_args=["--help"] if len(sys.argv) <= 1 else None,
        cli_settings_source=cli_settings_source,
    )


class _Config(_core.Config, toml_file=None):
    pass


class _New(_core.Server, alias_generator=None):
    """Create a new directory structure for Mahoraga.

    The default configuration may not be suitable for you. A small subset of
    the options can be tuned via command line (see below), while the rest live
    in the `mahoraga.toml` file created by this command. Once you've done with
    that file, you don't need to run this command again to create another
    directory. For details, run `mahoraga import -h`.
    """

    root: Annotated[
        pydantic_settings.CliPositionalArg[pydantic.NewPath],
        pydantic.Field(description="Root path of the new directory"),
    ]

    def cli_cmd(self) -> None:
        cfg = _Config(server=self)
        cfg_file = _setup(cfg, self.root)
        click.echo(f"Done. Please edit {cfg_file} before starting the server.")


class _Import(_core.Address):
    """Create a new directory structure from existing configuration.

    Mahoraga directory structure is not relocatable. Use this command to move
    your configuration files to another location, either on the same machine or
    not.

    Mahoraga directory structure follows semantic versioning. Directory
    created by Mahoraga version X.Y.Z (X>=1) is guaranteed to work under
    any version >=X.Y.Z,<X+1. Once updated to an uncompatible version,
    you have to create a new directory via `mahoraga new` and migrate your data
    by hand.
    """

    source: Annotated[
        pydantic_settings.CliPositionalArg[pydantic.FilePath],
        pydantic.Field(description="An existing mahoraga.toml file"),
        _core.Predicate("input_value.name == 'mahoraga.toml'"),
    ]
    destination: Annotated[
        pydantic_settings.CliPositionalArg[pydantic.NewPath],
        pydantic.Field(description="Root path of the new directory"),
    ]

    def cli_cmd(self) -> None:
        with contextlib.chdir(self.source.parent):
            cfg = _core.Config()
        cfg.server.host = self.host
        cfg.server.port = self.port
        cfg_file = _setup(cfg, self.destination)
        click.echo(f"Mahoraga root directory created at {cfg_file.parent}")


class _Run(pydantic.BaseModel, validate_default=True):
    """Start Mahoraga server.

    Before starting, make sure you've had all options in `mahoraga.toml`
    set properly. When the server is already running, changes in
    `mahoraga.toml` won't take effect until a restart.
    """

    root: Annotated[
        pydantic.DirectoryPath,
        pydantic.Field(description="Root path of a directory containing "
                                   "mahoraga.toml"),
        _core.Predicate("(input_value / 'mahoraga.toml').is_file()"),
    ] = pathlib.Path()

    def cli_cmd(self) -> None:
        _asgi.run(self.root)


class _Version(pydantic.BaseModel):
    """Show Mahoraga version and exit."""

    def cli_cmd(self) -> None:  # noqa: PLR6301
        click.echo(f"Mahoraga v{__version__}")


def _prog_name(arg0: str) -> str | None:
    return None if arg0.endswith("__main__.py") else arg0


def _setup(cfg: _core.Config, root: pathlib.Path) -> pathlib.Path:
    root.mkdir(parents=True)
    root = root.resolve(strict=True)
    for subdir in "channels", "log", "nginx", "repodata-cache":
        (root / subdir).mkdir()

    kwargs = cfg.model_dump()
    kwargs["server"]["root"] = root.as_posix()
    kwargs["unix"] = sys.platform.startswith(("darwin", "linux"))
    kwargs["upstream"]["python"] = [
        urllib.parse.unquote(str(url)) for url in kwargs["upstream"]["python"]
    ]

    env = jinja2.Environment(
        autoescape=True,
        loader=jinja2.PackageLoader(__name__, package_path=""),
    )
    cfg_file = root / "mahoraga.toml"
    for src, dst in [
        ("mahoraga.toml.jinja", cfg_file),
        ("mahoraga.conf.jinja", root / "nginx/mahoraga.conf"),
        ("nginx.conf.jinja", root / "nginx/nginx.conf"),
    ]:
        with dst.open("x", encoding="utf-8", newline="") as f:
            print(env.get_template(src).render(kwargs), file=f)
    return cfg_file


def _summary(docstring: str | None) -> str | None:
    return docstring.split(".", 1)[0] if docstring else None


class _Main(
    pydantic_settings.BaseSettings,
    cli_prog_name=_prog_name(sys.argv[0]),
    nested_model_default_partial_update=True,
    case_sensitive=True,
    cli_hide_none_type=True,
    cli_avoid_json=True,
    cli_enforce_required=True,
    cli_implicit_flags=True,
    cli_kebab_case=True,
):
    """Reverse proxy for Python mirrors."""

    new: Annotated[
        pydantic_settings.CliSubCommand[_New],
        pydantic.Field(description=_summary(_New.__doc__)),
    ]
    import_: Annotated[
        pydantic_settings.CliSubCommand[_Import],
        pydantic.Field(alias="import", description=_summary(_Import.__doc__)),
    ]
    run: Annotated[
        pydantic_settings.CliSubCommand[_Run],
        pydantic.Field(description=_summary(_Run.__doc__)),
    ]
    version: Annotated[
        pydantic_settings.CliSubCommand[_Version],
        pydantic.Field(description=_summary(_Version.__doc__)),
    ]

    def cli_cmd(self) -> None:
        pydantic_settings.CliApp.run_subcommand(self)



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["run"]

import asyncio
import collections
import concurrent.futures
import contextlib
import io
import logging.config
import logging.handlers
import multiprocessing as mp
import pathlib
import sys
import warnings
from typing import Any, override

import anysqlite
import hishel._core._spec  # pyright: ignore[reportMissingTypeStubs]
import httpx
import pooch.utils  # pyright: ignore[reportMissingTypeStubs]
import rich.console
import rich.logging
import uvicorn.config

from mahoraga import __version__, _conda, _core

from . import _app


def run() -> None:
    cfg = _core.Config()
    log_level = uvicorn.config.LOG_LEVELS[cfg.log.level]
    server_config = _ServerConfig(
        app=_app.make_app,
        host=str(cfg.server.host),
        port=cfg.server.port,
        loop="none",
        log_config={
            "version": 1,
            "handlers": {
                "default": {
                    "class": "logging.handlers.QueueHandler",
                    "queue": mp.get_context("spawn").Queue(),
                },
            },
            "root": {
                "handlers": ["default"],
                "level": log_level,
            },
            "disable_existing_loggers": False,
        },
        log_level=log_level,
        access_log=cfg.log.access,
        limit_concurrency=cfg.server.limit_concurrency,
        backlog=cfg.server.backlog,
        timeout_keep_alive=cfg.server.keep_alive,
        timeout_graceful_shutdown=cfg.server.timeout_graceful_shutdown or None,
        timeout_notify=3600,
        callback_notify=_conda.split_repo,
        factory=True,
    )
    server = uvicorn.Server(server_config)
    with server_config.listen:
        try:
            asyncio.run(
                _main(cfg, server),
                debug=cfg.log.level == "debug",
                loop_factory=cfg.loop_factory,
            )
        except KeyboardInterrupt:
            pass
        except SystemExit:
            if server.started:
                raise
            sys.exit(3)
        except BaseException as e:
            _logger.critical("ERROR", exc_info=e)
            raise SystemExit(server.started or 3) from e


async def _event_hook(request: httpx.Request) -> None:  # noqa: RUF029
    async def trace(event_name: str, _info: dict[str, Any]) -> None:  # noqa: RUF029
        _logger.debug("%s: %s", event_name, request.url)
    request.extensions["trace"] = trace


def _initializer(cfg: dict[str, Any]) -> None:
    logging.config.dictConfig(cfg)
    logging.captureWarnings(capture=True)
    pooch.utils.LOGGER = logging.getLogger("pooch")


async def _main(
    cfg: _core.Config,
    server: uvicorn.Server,
) -> None:
    if cfg.log.level == "debug":
        event_hooks = {"request": [_event_hook]}
    else:
        event_hooks = None

    log_config = server.config.log_config
    if not isinstance(log_config, dict):
        _core.unreachable()

    async with contextlib.AsyncExitStack() as stack:
        _core.context.set({
            "config": cfg,
            "httpx_client": await stack.enter_async_context(
                _core.AsyncClient(
                    headers={"User-Agent": f"mahoraga/{__version__}"},
                    timeout=httpx.Timeout(15, read=60, write=60),
                    follow_redirects=False,
                    limits=httpx.Limits(keepalive_expiry=cfg.server.keep_alive),
                    event_hooks=event_hooks,
                    storage=hishel.AsyncSqliteStorage(
                        connection=await anysqlite.connect(":memory:"),
                        default_ttl=600.,
                    ),
                ),
            ),
            "locks": _core.WeakValueDictionary(),
            "process_pool": stack.enter_context(
                concurrent.futures.ProcessPoolExecutor(
                    initializer=_initializer,
                    initargs=(log_config,),
                    max_tasks_per_child=1000,
                ),
            ),
            "statistics_concurrent_requests": collections.Counter(),
        })
        await server.serve()


class _RotatingFileHandler(logging.handlers.RotatingFileHandler):
    def _open(self) -> io.TextIOWrapper:
        if self.mode != "a":
            _core.unreachable()
        return pathlib.Path(self.baseFilename).open(
            mode="a",
            encoding=self.encoding,
            errors=self.errors,
            newline="",
        )


class _ServerConfig(uvicorn.Config):
    @override
    def configure_logging(self) -> None:
        super().configure_logging()
        logging.getLogger("hishel").setLevel(logging.DEBUG)
        logging.getLogger("hishel.core.spec").addFilter(
            lambda rec: rec.msg != "Storing response in cache",
        )
        logging.getLogger("hishel.integrations.clients").addFilter(
            lambda rec: rec.msg != "Handling state: IdleClient",
        )
        if self.access_log:
            logging.getLogger("uvicorn.access").setLevel(logging.INFO)
        if self.log_level == logging.DEBUG:
            warnings.simplefilter("always", ResourceWarning)
        logging.captureWarnings(capture=True)
        pooch.utils.LOGGER = logging.getLogger("pooch")

        root = logging.getLogger()
        [old] = root.handlers
        if not isinstance(old, logging.handlers.QueueHandler):
            _core.unreachable()
        root.removeHandler(old)

        log_dir = pathlib.Path("log")
        log_dir.mkdir(exist_ok=True)
        log_dir = log_dir.resolve(strict=True)
        new = _RotatingFileHandler(
            log_dir / "mahoraga.log",
            maxBytes=20000 * 81,  # lines * chars
            backupCount=10,
            encoding="utf-8",
        )
        fmt = logging.Formatter(
            "[{asctime}] {levelname:8} {message}",
            datefmt="%Y-%m-%d %X",
            style="{",
        )
        new.setFormatter(fmt)
        root.addHandler(new)

        if isinstance(sys.stdout, io.TextIOBase) and sys.stdout.isatty():
            if rich.console.detect_legacy_windows():
                new = logging.StreamHandler(sys.stdout)
                new.setFormatter(fmt)
            else:
                new = rich.logging.RichHandler(log_time_format="[%Y-%m-%d %X]")
            root.addHandler(new)

        self.listen = logging.handlers.QueueListener(old.queue, *root.handlers)


hishel._core._spec.get_heuristic_freshness = lambda response: 600  # noqa: ARG005, SLF001  # ty: ignore[invalid-assignment]
_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "AsyncClient",
    "Context",
    "Statistics",
    "WeakValueDictionary",
    "cache_action",
    "cached_or_locked",
    "schedule_exit",
]

import asyncio
import collections
import contextlib
import contextvars
import dataclasses
import time
import weakref
from typing import TYPE_CHECKING, Any, TypedDict, overload, override

import anyio
import hishel.httpx
import httpx
import httpx_aiohttp
import pydantic_settings
from rattler.networking.fetch_repo_data import CacheAction

from mahoraga import _core

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Awaitable
    from concurrent.futures import ProcessPoolExecutor

    from _typeshed import StrPath


class AsyncClient(
    httpx_aiohttp.HttpxAiohttpClient,
    hishel.httpx.AsyncCacheClient,
):
    @override
    def _init_transport(
        self,
        *args: Any,
        **kwargs: object,
    ) -> httpx.AsyncBaseTransport:
        next_transport = super()._init_transport(*args, **kwargs)
        return _AsyncCacheTransport(
            next_transport=next_transport,
            storage=self.storage,
            policy=self.policy,
        )

    @override
    def _transport_for_url(self, url: httpx.URL) -> httpx.AsyncBaseTransport:
        t = super()._transport_for_url(url)
        if isinstance(t, hishel.httpx.AsyncCacheTransport):
            match cache_action.get():
                case "no-cache":
                    return t.next_transport
                case "force-cache-only" | "use-cache-only":
                    return hishel.httpx.AsyncCacheTransport(
                        _not_implemented,
                        t.storage,
                        t._cache_proxy.policy,  # noqa: SLF001  # pyright: ignore[reportPrivateUsage]
                    )
                case "cache-or-fetch":
                    pass
        return t

    @override
    @contextlib.asynccontextmanager
    async def stream(
        self,
        method: str,
        url: httpx.URL | str,
        **kwargs: Any,
    ) -> AsyncGenerator[httpx.Response]:
        url = httpx.URL(url)
        h = url.host
        if h.endswith((
            "anaconda.org",
            "github.com",
            "prefix.dev",
            "pypi.org",
        )):
            kwargs["follow_redirects"] = True
        cm = super().stream(method, url, **kwargs)
        ctx = _core.context.get()
        concurrent_requests = ctx["statistics_concurrent_requests"]
        concurrent_requests[h] += 1
        async with contextlib.AsyncExitStack() as stack:
            tic = time.monotonic()
            try:
                yield await stack.enter_async_context(cm)
            finally:
                toc = time.monotonic()
                concurrent_requests[h] -= 1
                if seconds := round(toc - tic):
                    schedule_exit(stack)
                    async with ctx["locks"]["statistics.json"]:
                        s = Statistics()
                        s.total_seconds[h] += seconds
                        await _json.write_text(
                            s.model_dump_json(exclude=_exclude),
                            encoding="utf-8",
                        )


def schedule_exit(stack: contextlib.AsyncExitStack) -> None:
    task = asyncio.create_task(stack.pop_all().aclose())
    stack.push_async_callback(lambda: task)


class Statistics(pydantic_settings.BaseSettings, json_file_encoding="utf-8"):
    backup_servers: set[str] = set()
    concurrent_requests: collections.Counter[str] = collections.Counter()
    total_seconds: collections.Counter[str] = collections.Counter()

    def key(self, url: str) -> tuple[bool, int, int]:
        h = httpx.URL(url).host
        return (
            h in self.backup_servers,
            self.concurrent_requests[h],
            self.total_seconds[h],
        )

    @override
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[pydantic_settings.BaseSettings],
        init_settings: pydantic_settings.PydanticBaseSettingsSource,
        env_settings: pydantic_settings.PydanticBaseSettingsSource,
        dotenv_settings: pydantic_settings.PydanticBaseSettingsSource,
        file_secret_settings: pydantic_settings.PydanticBaseSettingsSource,
    ) -> tuple[pydantic_settings.PydanticBaseSettingsSource, ...]:
        ctx = _core.context.get()
        json_settings = pydantic_settings.JsonConfigSettingsSource(
            settings_cls, "statistics.json")
        json_settings.init_kwargs.update(  # pyright: ignore[reportUnknownMemberType]
            backup_servers=ctx["config"].upstream.backup,
            concurrent_requests=ctx["statistics_concurrent_requests"],
        )
        return (json_settings,)


class WeakValueDictionary(weakref.WeakValueDictionary[str, asyncio.Lock]):
    @override
    def __getitem__(self, key: str) -> asyncio.Lock:
        try:
            return super().__getitem__(key)
        except KeyError:
            self[key] = value = asyncio.Lock()
            return value


@overload
def cached_or_locked(
    cache_location: StrPath,
    stack: None = ...,
) -> contextlib.AbstractAsyncContextManager[bool]: ...

@overload
async def cached_or_locked(
    cache_location: StrPath,
    stack: contextlib.AsyncExitStack,
) -> bool: ...


def cached_or_locked(
    cache_location: StrPath,
    stack: contextlib.AsyncExitStack | None = None,
) -> Awaitable[bool] | contextlib.AbstractAsyncContextManager[bool]:
    cm = _cached_or_locked(cache_location)
    return stack.enter_async_context(cm) if stack else cm


@contextlib.asynccontextmanager
async def _cached_or_locked(cache_location: StrPath) -> AsyncGenerator[bool]:
    ctx = _core.context.get()
    async with ctx["locks"][str(cache_location)]:
        if not await anyio.Path(cache_location).is_file():
            yield False
            return
    yield True


class _AsyncCacheProxy(hishel.AsyncCacheProxy):
    @override
    async def _get_key_for_request(self, request: hishel.Request) -> str:
        if headers := request.headers.get_list("accept"):
            for header in headers:
                if header.startswith("application/vnd.pypi.simple.v1+"):
                    project = request.url.rsplit("/", 2)[1]
                    return f"{project}|{header}"
        return await super()._get_key_for_request(request)

    @override
    async def _handle_idle_state(
        self,
        state: hishel.IdleClient,
        request: hishel.Request,
    ) -> hishel.AnyState:
        stored_entries = [
            dataclasses.replace(
                pair,
                request=dataclasses.replace(pair.request, url=request.url),
            )
            for pair in await self.storage.get_entries(
                await self._get_key_for_request(request),
            )
        ]
        return state.next(request, stored_entries)


class _AsyncCacheTransport(hishel.httpx.AsyncCacheTransport):
    @override
    def __init__(
        self,
        next_transport: httpx.AsyncBaseTransport,
        storage: hishel.AsyncBaseStorage | None = None,
        policy: hishel.CachePolicy | None = None,
    ) -> None:
        self.next_transport = next_transport
        self._cache_proxy = _AsyncCacheProxy(
            request_sender=self.request_sender,
            storage=storage,
            policy=policy,
        )
        self.storage = self._cache_proxy.storage


class _Context(TypedDict):
    config: _core.Config
    httpx_client: AsyncClient
    locks: WeakValueDictionary
    process_pool: ProcessPoolExecutor
    statistics_concurrent_requests: collections.Counter[str]


Context = contextvars.ContextVar[_Context]
cache_action = contextvars.ContextVar[CacheAction](
    "cache_action",
    default="no-cache",
)
_exclude = {"backup_servers", "concurrent_requests"}
_json = anyio.Path("statistics.json")
_not_implemented = httpx.AsyncBaseTransport()



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "JLAPHeaders",
    "PackageRecord",
    "RepodataHeaders",
    "RunExports",
    "Shard",
    "ShardedRepodata",
]

from typing import TYPE_CHECKING, Any, TypedDict

import pydantic

if TYPE_CHECKING:
    from rattler.platform import PlatformLiteral


class _RepodataHeaders(pydantic.BaseModel, extra="ignore"):
    if_none_match: str | None = None


class JLAPHeaders(_RepodataHeaders):
    range: str | None = None


class PackageRecord(pydantic.BaseModel, extra="allow"):
    md5: str | None = None
    sha256: str | None = None


class RepodataHeaders(_RepodataHeaders):
    accept_encoding: str | None = None
    if_modified_since: str | None = None


class RunExports(
    TypedDict("_RunExports", {"packages.conda": dict[str, dict[str, Any]]}),
):
    packages: dict[str, dict[str, Any]]


class Shard(RunExports):
    removed: list[str]


class _ShardedSubdirInfo(TypedDict):
    base_url: str
    shards_base_url: str
    subdir: PlatformLiteral


class ShardedRepodata(TypedDict):
    info: _ShardedSubdirInfo
    shards: dict[str, bytes]



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["extract_from_tarball", "get_npm_file", "urls"]

import asyncio
import base64
import logging
import mimetypes
import pathlib
import posixpath
import shutil
import tarfile
from typing import TYPE_CHECKING, Literal

import fastapi.responses

from mahoraga import _core, _jsdelivr

if TYPE_CHECKING:
    from contextlib import AsyncExitStack


async def extract_from_tarball(
    tarball: pathlib.Path,
    member: str,
    cache_location: pathlib.Path,
) -> fastapi.Response | None:
    ctx = _core.context.get()
    loop = asyncio.get_running_loop()
    async with ctx["locks"][str(tarball)]:
        try:
            f = await loop.run_in_executor(
                None,
                tarfile.TarFile.bz2open,
                tarball,
            )
        except OSError:
            return None
    try:
        member = (
            "xbuildenv/pyodide-root/dist/"
            if tarball.name.startswith("xbuildenv")
            else "pyodide/"
        ) + member
        return await loop.run_in_executor(
            None,
            _extract_from_tarball,
            tarball,
            member,
            cache_location,
            f,
        )
    finally:
        await loop.run_in_executor(None, f.close)


async def get_npm_file(
    url: str,
    package: str,
    path: str,
    cache_location: pathlib.Path,
    stack: AsyncExitStack,
) -> fastapi.Response:
    if package.startswith("pyodide@"):
        for name in _pyodide_packages(path):
            tarball = pathlib.Path("pyodide", f"{name}-{package[8:]}.tar.bz2")
            if response := await extract_from_tarball(
                tarball,
                path,
                cache_location,
            ):
                return response
    metadata = await _jsdelivr.Metadata.fetch(
        f"npm/{package}.json",
        url=url,
        params={"structure": "flat"},
    )
    for file in metadata.files:
        if file["name"].lstrip("/") == path:
            break
    else:
        return fastapi.Response(status_code=404)
    media_type, _ = mimetypes.guess_type(path)
    return await _core.stream(
        urls("npm", package, path),
        media_type=media_type,
        stack=stack,
        cache_location=cache_location,
        sha256=base64.b64decode(file["hash"]),
        size=file["size"],
    )


def urls(*paths: str) -> list[str]:
    return [
        posixpath.join(str(url), *paths)
        for url in _core.context.get()["config"].upstream.pyodide
    ]


def _extract_from_tarball(
    tarball: pathlib.Path,
    member: str,
    cache_location: pathlib.Path,
    f: tarfile.TarFile,
) -> fastapi.Response | None:
    try:
        info = f.getmember(member)
    except KeyError:
        _logger.warning("Invalid tarball %s", tarball, exc_info=True)
        return None
    if f2 := f.extractfile(info):
        with f2:
            cache_location.parent.mkdir(parents=True, exist_ok=True)
            with cache_location.open("xb") as f3:
                f3.truncate(info.size)
                shutil.copyfileobj(f2, f3)
        return fastapi.responses.FileResponse(cache_location)
    _logger.warning("Invalid tarball %s", tarball)
    return None


def _pyodide_packages(
    path: str,
) -> tuple[Literal["pyodide", "pyodide-core", "xbuildenv"], ...]:
    match path:
        case (
            "pyodide.asm.js"
            | "pyodide.asm.wasm"
            | "pyodide.d.ts"
            | "pyodide.js"
            | "pyodide.mjs"
            | "pyodide-lock.json"
            | "python_stdlib.zip"
        ):
            return ("pyodide-core", "xbuildenv", "pyodide")
        case "ffi.d.ts" | "package.json":
            return ("pyodide-core", "pyodide")
        case "pyodide.js.map" | "pyodide.mjs.map":
            return ("xbuildenv", "pyodide")
        case "console.html":
            return ("pyodide",)
        case _:
            return ()


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "APIRoute",
    "Address",
    "AsyncClient",
    "Config",
    "GitHubRelease",
    "NPMBase",
    "Predicate",
    "Response",
    "Server",
    "Statistics",
    "StreamingResponse",
    "WeakValueDictionary",
    "cache_action",
    "cached_or_locked",
    "context",
    "get",
    "hourly",
    "immutable",
    "load_balance",
    "schedule_exit",
    "stream",
    "unreachable",
]

from typing import NoReturn

from ._config import Address, Config, Predicate, Server
from ._context import (
    AsyncClient,
    Context,
    Statistics,
    WeakValueDictionary,
    cache_action,
    cached_or_locked,
    schedule_exit,
)
from ._metadata import GitHubRelease, NPMBase
from ._stream import (
    APIRoute,
    Response,
    StreamingResponse,
    get,
    hourly,
    immutable,
    load_balance,
    stream,
)

context = Context("context")


def unreachable(message: str = "Unreachable") -> NoReturn:
    raise AssertionError(message)



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import asyncio
import collections
import contextlib
import contextvars
import http
import posixpath
from typing import Annotated, Literal

import fastapi
import kiss_headers

from mahoraga import _core

from . import _models  # noqa: TC001

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/{project}/")
async def get_pypi_project(
    project: _models.NormalizedName,
    accept: Annotated[
        str | None,
        fastapi.Header(
            description="See [PEP 691](https://peps.python.org/pep-0691/)",
        ),
    ] = None,
    *,
    micropip: Annotated[bool, fastapi.Query()] = False,
) -> fastapi.Response:
    ctx = contextvars.copy_context()
    match ctx[_core.context]:
        case {"config": config, "locks": locks}:
            pass
        case _:
            _core.unreachable()
    if micropip or not config.upstream.pypi.json_:
        media_type = "application/vnd.pypi.simple.v1+html"
    else:
        media_type = _decide_content_type(accept)
    if media_type == "application/vnd.pypi.simple.v1+json":
        urls = [
            posixpath.join(str(url), "simple", project) + "/"
            for url in config.upstream.pypi.json_
        ]
    else:
        urls = [
            posixpath.join(str(url), "simple", project) + "/"
            for url in config.upstream.pypi.all()
        ]
    ctx.run(_core.cache_action.set, "cache-or-fetch")
    async with contextlib.AsyncExitStack() as stack:
        await stack.enter_async_context(locks[f"{project}|{media_type}"])
        return await asyncio.create_task(
            _core.stream(
                urls,
                headers={"Accept": media_type},
                media_type=media_type,
                stack=stack,
            ),
            context=ctx,
        )


def _decide_content_type(accept: str | None) -> Literal[
    "application/vnd.pypi.simple.v1+json",
    "application/vnd.pypi.simple.v1+html",
    "text/html",
]:
    if not accept:
        return "application/vnd.pypi.simple.v1+html"
    g: collections.defaultdict[float | None, set[str | None]]
    g = collections.defaultdict(set)
    match kiss_headers.get_polymorphic(
        kiss_headers.parse_it("Accept: " + accept),
        kiss_headers.Accept,
    ):
        case list(headers):
            for h in headers:
                g[h.get_qualifier()].add(h.get_mime())
        case None:
            _core.unreachable()
        case h:
            g[h.get_qualifier()].add(h.get_mime())
    for _, v in sorted(g.items(), reverse=True):
        if (
            "application/vnd.pypi.simple.v1+json" in v
            or "application/vnd.pypi.simple.latest+json" in v
        ):
            return "application/vnd.pypi.simple.v1+json"
        if (
            "application/vnd.pypi.simple.v1+html" in v
            or "application/vnd.pypi.simple.latest+html" in v
            or "application/*" in v
            or "*/*" in v
        ):
            return "application/vnd.pypi.simple.v1+html"
        if "text/html" in v or "text/*" in v:
            return "text/html"
    raise fastapi.HTTPException(http.HTTPStatus.NOT_ACCEPTABLE)
